/*
 * I2S Interface
 *   - Initializes HW Codec
 *   - Manages 8 KHz I2S stream to codec
 *   - Provides circular buffer and semaphore protected access routines
 *
 * Copyright (c) 2023 Dan Julio
 *
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2, as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */
#include <string.h>
#include "audio_hal.h"
#include "audio_task.h"
#include "esp_system.h"
#include "esp_log.h"
#include "freertos/FreeRTOS.h"
#include "freertos/task.h"
#include "freertos/semphr.h"
#include "freertos/queue.h"
#include "driver/i2s.h"
#include "gain.h"
#include "spandsp.h"


//
// Local constants
//

// SAMPLE RATE
#define AUDIO_SAMPLE_RATE 8000

// Number of samples to read/write to the I2S subsystem at a time (bytes = 4x)
//   Multiple is in mSec.  This should be the FreeRTOS scheduler period
#define I2S_SAMPLES (10 * AUDIO_SAMPLE_RATE / 1000)

// Number of samples in our circular buffers
//   Optimized to be a set of samples or two larger than the most entries used during operation
//   as measured with AUDIO_PRINT_BUF_INFO defined
#define BUF_SAMPLES (8 * I2S_SAMPLES)

// Maximum amount of data to read from the I2S driver to prevent it from overflowing
// by reading more than one full I2S_SAMPLES if available.
#define MAX_READ_NUM_SAMPLES 3

// Notification decode macro
#define Notification(var, mask) ((var & mask) == mask)



//
// Variables
//

static const char* TAG = "audio_task";

TaskHandle_t task_handle_audio;

// State
static bool audio_enabled = false;

// Incoming audio circular buffer
static int16_t rx_buf[BUF_SAMPLES];
static int rx_buf_put;
static int rx_buf_pop;
static int rx_buf_count;
static SemaphoreHandle_t rx_buf_mutex;

// I2S buffer (2 entries per sample for both L+R channels)
static int16_t i2s_rx_buf[MAX_READ_NUM_SAMPLES*2*I2S_SAMPLES];

// DC restore state
static dc_restore_state_t dc_restore_state;

// I2S event queue
static QueueHandle_t i2s_event_queue;



//
// Forward declarations
//
static void _audioInitI2S();
static bool _audioInitCodec();
static void _audioInitBuffers();
static void _audioHandleNotifications();
static int _audioGetRx(int16_t* buf, int len);
static void _audioPutRx(int len, int16_t* i2s_rxP);

//
// API
//
void audio_task(void* args)
{
	int i;
	size_t bytes_read;
	i2s_event_t i2s_evt;
	
	
	ESP_LOGI(TAG, "Start task");
	
	// Create circular buffer access mutex
	rx_buf_mutex = xSemaphoreCreateMutex();
	_audioInitBuffers();
	
	// configure i2s
	_audioInitI2S();
    
    // configure codec
    if (_audioInitCodec()) {
    	ESP_LOGI(TAG, "Codec initialized");
    } else {
    	ESP_LOGE(TAG, "Codec init failed");
    	vTaskDelete(NULL);
    }
    (void) i2s_stop(I2S_NUM_0);
    
    while (true) {
    	if (!audio_enabled) {
    		// Do nothing but wait to be enabled
    		_audioHandleNotifications();
    		vTaskDelay(pdMS_TO_TICKS(10));
    	} else {    	
    		// Start the codec
    		(void) audio_hal_ctrl_codec(AUDIO_HAL_CODEC_MODE_BOTH, AUDIO_HAL_CTRL_START);
    		
    		// Start I2S
    		(void) i2s_start(I2S_NUM_0);
    	
		   	while (audio_enabled) {
		   		while (xQueueReceive(i2s_event_queue, &i2s_evt, 0) && audio_enabled) {
					if (i2s_evt.type == I2S_EVENT_RX_DONE) {
						// Set timeout to 0 to get whatever is available without blocking.
						// Read up to MAX_READ_NUM_SAMPLES complete sets of samples to try to prevent driver overflows.
				    	(void) i2s_read(I2S_NUM_0, (void*) i2s_rx_buf, MAX_READ_NUM_SAMPLES * I2S_SAMPLES * 4, &bytes_read, 0);
						
						// DC restoration to remove any DC offsets because they may interfere with tone detection
						for (i=0; i<(bytes_read/2); i+=2) {
							i2s_rx_buf[i] = dc_restore(&dc_restore_state, i2s_rx_buf[i]);
						}

				    	
				    	// Store rx data (number of samples = 1/4 bytes read)
				    	_audioPutRx(bytes_read/4, i2s_rx_buf);
			    	} else if (i2s_evt.type == I2S_EVENT_RX_Q_OVF) {
			    		ESP_LOGE(TAG, "I2S RX OVFL");
			    	} else if (i2s_evt.type == I2S_EVENT_DMA_ERROR) {
			    		ESP_LOGE(TAG, "I2S DMA ERROR");
			    	}
			    	
			    	_audioHandleNotifications();
		   		}
		   		
				vTaskDelay(pdMS_TO_TICKS(10));
			}
			
			(void) i2s_stop(I2S_NUM_0);
			(void) audio_hal_ctrl_codec(AUDIO_HAL_CODEC_MODE_BOTH, AUDIO_HAL_CTRL_STOP);

			// Reset the buffers
			_audioInitBuffers();
		}
	}
}


int audioGetRxCount()
{
	int n;
	
	xSemaphoreTake(rx_buf_mutex, portMAX_DELAY);
	n = rx_buf_count;
	xSemaphoreGive(rx_buf_mutex);
	
	return n;
}


int audioGetToneRx(int16_t* buf, int len)
{
	if (audio_enabled) {
		return _audioGetRx(buf, len);
	} else {
		for (int i=0; i<len; i++) buf[i] = 0;
		return len;
	}
}



//
// Internal functions
//
static void _audioInitI2S()
{
	i2s_config_t i2s_config = {
        .mode = I2S_MODE_MASTER | I2S_MODE_RX | I2S_MODE_TX,
        .sample_rate =  AUDIO_SAMPLE_RATE,
        .bits_per_sample = I2S_BITS_PER_SAMPLE_16BIT,
        .communication_format = I2S_COMM_FORMAT_STAND_I2S,
        .intr_alloc_flags = ESP_INTR_FLAG_LEVEL2 | ESP_INTR_FLAG_IRAM,
		.channel_format = I2S_CHANNEL_FMT_RIGHT_LEFT,
        .dma_buf_count = 3,  // Three buffers to help prevent driver underflow/overflow conditions
        .dma_buf_len = I2S_SAMPLES,
        .use_apll = 1,
        .tx_desc_auto_clear = 1,
        .mclk_multiple = 256
	};
	i2s_pin_config_t i2s_pin_config = {
		.bck_io_num = GPIO_NUM_27,
		.ws_io_num = GPIO_NUM_25,
		.data_out_num = GPIO_NUM_26,
		.data_in_num = GPIO_NUM_35
	};
	
    // install i2s driver
    i2s_driver_install(I2S_NUM_0, &i2s_config, 8, &i2s_event_queue);
    i2s_set_pin(I2S_NUM_0, &i2s_pin_config);
}


static bool _audioInitCodec()
{
	audio_hal_codec_config_t codec_config = AUDIO_HAL_ES8388_DEFAULT();
	
	if (!audio_hal_init(&codec_config, AUDIO_CODEC_ES8388)) {
		return false;
	}
	
	// Set initial volume
	if (!gainSetCodec(GAIN_TYPE_MIC, GAIN_APP_MIC_NOM_DB)) {
		return false;
	}
	
	if (!gainSetCodec(GAIN_TYPE_SPK, GAIN_APP_SPK_MIN_DB)) {
		return false;
	}
	
	return true;
}


static void _audioInitBuffers()
{
	xSemaphoreTake(rx_buf_mutex, portMAX_DELAY);
	rx_buf_put = 0;
	rx_buf_pop = 0;
	rx_buf_count = 0;
	xSemaphoreGive(rx_buf_mutex);
}


static void _audioHandleNotifications()
{
	uint32_t notification_value = 0;
	
	// Handle notifications (clear them upon reading)
	if (xTaskNotifyWait(0x00, 0xFFFFFFFF, &notification_value, 0)) {
		if (Notification(notification_value, AUDIO_NOTIFY_DISABLE_MASK)) {
			if (audio_enabled != false) {
				ESP_LOGI(TAG, "Disable stream");
				audio_enabled = false;
			}
		}
		
		if (Notification(notification_value, AUDIO_NOTIFY_ENABLE_MASK)) {
			if (audio_enabled != true) {
				ESP_LOGI(TAG, "Enable stream (8k)");
				audio_enabled = true;
				
				// Initialize zero DC restoration machine
				dc_restore_init(&dc_restore_state);
			}
		}
	}
}


static int _audioGetRx(int16_t* buf, int len)
{
	int i;
	int read_len;
	
	xSemaphoreTake(rx_buf_mutex, portMAX_DELAY);
	read_len = (len > rx_buf_count) ? rx_buf_count : len;
	
	for (i=0; i<read_len; i++) {
		*(buf+i) = rx_buf[rx_buf_pop++];
		if (rx_buf_pop >= BUF_SAMPLES) rx_buf_pop = 0;
	}
	
	rx_buf_count -= read_len;
	xSemaphoreGive(rx_buf_mutex);
	
	// Fill remaining with zero if necessary
	if (len > read_len) {
		for (i=read_len; i<len; i++) {
			*(buf+i) = 0;
		}
	}
	
	return read_len;
}


static void _audioPutRx(int len, int16_t* i2s_rxP)
{
	int i;
	
	// Quickly load the processed data
	xSemaphoreTake(rx_buf_mutex, portMAX_DELAY);
	for (i=0; i<len; i++) {
		rx_buf[rx_buf_put++] = *i2s_rxP;
		i2s_rxP += 2;                   // Skip other channel
		if (rx_buf_put >= BUF_SAMPLES) rx_buf_put = 0;
	}
	
	rx_buf_count += len;
	if (rx_buf_count > BUF_SAMPLES) {
		rx_buf_count = rx_buf_count % BUF_SAMPLES;
		ESP_LOGE(TAG, "RX FIFO overflow");
	}
	xSemaphoreGive(rx_buf_mutex);
}
