/*
 * Caller ID audio and decode task
 */
#include "esp_system.h"
#include "esp_log.h"
#include "freertos/FreeRTOS.h"
#include "freertos/queue.h"
#include "audio_task.h"
#include "cid_task.h"
#include "gui_task.h"
#include "spandsp.h"



//
// Constants
//

// One of the adsi.h standard enum values
//    ADSI_STANDARD_CLASS
//    ADSI_STANDARD_CLIP
//    ADSI_STANDARD_CLIP_DTMF
#define CALLER_ID_STANDARD        ADSI_STANDARD_CLASS

// Evaluation interval
#define ADSI_EVAL_MSEC            10

// ADSI Buffer size - calculated to be one evaluation interval
#define ADSI_BUF_LEN              (8000 * ADSI_EVAL_MSEC / 1000)

// Notification decode macro
#define Notification(var, mask) ((var & mask) == mask)



//
// Global Variables
//
static const char* TAG = "cid_task";

TaskHandle_t task_handle_cid;

// For callback (cribbed from adsi test code)
static int current_standard = ADSI_STANDARD_CLASS;

static bool process_audio = false;

// ADSI Decoder
static adsi_rx_state_t *rx_adsi;
static int16_t adsi_rx_buf[ADSI_BUF_LEN];



//
// Forward declarations
//
static void adsi_msg_cb(void *user_data, const uint8_t *msg, int len);
static void format_sdmf_number(const char* src, int len, char* dst);
static void format_mdmf_number(const char* src, int len, char* dst);


//
// API
//
void cid_task(void* args)
{
	int cur_samples_in_rx;
	int samples_to_analyze;
	uint32_t notification_value;
	
	ESP_LOGI(TAG, "Task start");
	
	
	while (1) {
		if (process_audio) {
			cur_samples_in_rx = audioGetRxCount();
			while (cur_samples_in_rx > 0) {
				samples_to_analyze = (cur_samples_in_rx >= ADSI_BUF_LEN) ? ADSI_BUF_LEN : cur_samples_in_rx;
				audioGetToneRx(adsi_rx_buf, samples_to_analyze);
				adsi_rx(rx_adsi, adsi_rx_buf, samples_to_analyze);
				cur_samples_in_rx -= samples_to_analyze;
			}
		}
		
		// Handle notifications (clear them upon reading)
		notification_value = 0;
		if (xTaskNotifyWait(0x00, 0xFFFFFFFF, &notification_value, 0)) {
			if (Notification(notification_value, CID_NOTIFY_START_SAMPLE)) {
				ESP_LOGI(TAG, "Start");
				process_audio = true;
				
				// (Re)initialize the decoder
				rx_adsi = adsi_rx_init(NULL, current_standard, adsi_msg_cb, NULL);
				
				// Start audio
				xTaskNotify(task_handle_audio, AUDIO_NOTIFY_ENABLE_MASK, eSetBits);
			}
			if (Notification(notification_value, CID_NOTIFY_STOP_SAMPLE)) {
				ESP_LOGI(TAG, "Stop");
				process_audio = false;
				
				// Disable audio
				xTaskNotify(task_handle_audio, AUDIO_NOTIFY_DISABLE_MASK, eSetBits);
				
				// Release decoder resources
				(void) adsi_rx_free(rx_adsi);
			}
		}
		
		vTaskDelay(pdMS_TO_TICKS(ADSI_EVAL_MSEC));
	}
}


//
// Internal routines
//
static void adsi_msg_cb(void *user_data, const uint8_t *msg, int len)
{
	int pos = -1;
    uint8_t field_type;
    const uint8_t *field_body;
    int field_len;
    int message_type = -1;
    char body[256];
    
    do {
    	pos = adsi_next_field(rx_adsi, msg, len, pos, &field_type, &field_body, &field_len);
    	if (pos > 0) {
    		if (field_body == NULL) {
    			// Message type - should see this first
    			ESP_LOGI(TAG, "Message type 0x%x", field_type);
    			message_type = field_type;
    		} else {
    			// Message field - Only CLASS/Bellcore at this time
    			memcpy(body, field_body, field_len);
                body[field_len] = '\0';
    			switch (message_type) {
    				case CLASS_SDMF_CALLERID:
                		format_sdmf_number((const char *) field_body, field_len, body);
    					gui_set_secondary_msg((const char*) body, CID_MESSAGE_TIMEOUT_SEC);
    					xTaskNotify(task_handle_gui, GUI_NOTIFY_SECONDARY_MESSAGE, eSetBits);
    					break;
    					
    				case CLASS_MDMF_CALLERID:
    					switch (field_type) {
    						case MCLASS_DATETIME:
    							break;
    						case MCLASS_CALLER_NUMBER:
    							format_mdmf_number((const char *) field_body, field_len, body);
    							gui_set_secondary_msg((const char*) body, CID_MESSAGE_TIMEOUT_SEC);
    							xTaskNotify(task_handle_gui, GUI_NOTIFY_SECONDARY_MESSAGE, eSetBits);
    							break;
    						case MCLASS_DIALED_NUMBER:
    							break;
    						case MCLASS_ABSENCE1:
    							break;
    						case MCLASS_REDIRECT:
    							break;
    						case MCLASS_QUALIFIER:
    							break;
    						case MCLASS_CALLER_NAME:
    							gui_set_primary_msg((const char*) body, CID_MESSAGE_TIMEOUT_SEC);
    							xTaskNotify(task_handle_gui, GUI_NOTIFY_PRIMARY_MESSAGE, eSetBits);
    							break;
    						case MCLASS_ABSENCE2:
    							break;
    					}
    					break;
    					
    				case CLASS_SDMF_MSG_WAITING:
    					break;
    					
    				case CLASS_MDMF_MSG_WAITING:
    					switch (field_type) {
    						case MCLASS_VISUAL_INDICATOR:
    							break;
    					}
    					break;
    			}
    			ESP_LOGI(TAG, "  Field Type: 0x%x, len %d, '%s'", field_type, field_len, body);
    		}
    	}
    } while (pos > 0);
    
    if (pos < -1) {
    	ESP_LOGE(TAG, "Bad message contents");
    }
}


static void format_sdmf_number(const char* src, int len, char* dst)
{
	// Skip the first part of the string - "MMDDHHmm+" 
	format_mdmf_number(src + 9, len - 9, dst);
}


static void format_mdmf_number(const char* src, int len, char* dst)
{
	if (len == 11) {
		// Strip off leading digit (probably a '1')
		src++;
		len -= 1;
	}
	
	if (len == 10) {
		// (NNN) NNN-NNNN
		*(dst++) = '(';
		
		memcpy(dst, src, 3);
		src += 3;
		dst += 3;
		
		*(dst++) = ')';
		*(dst++) = ' ';
		
		memcpy(dst, src, 3);
		src += 3;
		dst += 3;
		
		*(dst++) = '-';
		
		memcpy(dst, src, 4);
		dst += 4;
		*(dst) = 0;
	} else {
		// Just raw
		memcpy(dst, src, len);
		*(dst+len) = 0;
	}
}
