/*
 * Ring detection
 */
#include "driver/gpio.h"
#include "esp_system.h"
#include "esp_log.h"
#include "freertos/FreeRTOS.h"
#include "cid_task.h"
#include "ring_task.h"


//
// Constants
//

// Ring state
#define RING_IDLE  0
#define RING_PULSE 1
#define RING_WAIT  2

// Maximum ring pulse-to-pulse delay
#define RING_MAX_PULSE_MSEC 100

// Maximum period between rings to determine no more are coming
#define RING_MAX_WAIT_MSEC  5000



//
// Global Variables
//
static const char* TAG = "ring_task";

TaskHandle_t task_handle_ring;





//
// API
//
void ring_task(void* args)
{
	bool cur_ring_pin;
	bool ring_detected;
	bool prev_ring_pin = false;
	bool cid_enabled = false;
	int ring_state = RING_IDLE;
	int ring_counts = 0;
	int ring_pulse_counts = 0;
	int ring_pulse_avg_sum = 0;
	int ring_timer;
	int64_t cur_ring_usec;
	int64_t prev_ring_usec;
	
	// Setup the input
	gpio_reset_pin(RING_PIN);
	gpio_set_direction(RING_PIN, GPIO_MODE_INPUT);
	
	while (1) {
		// Get the current ring pin state
		cur_ring_pin = (gpio_get_level(RING_PIN) == 1);
		ring_detected = cur_ring_pin && !prev_ring_pin;
		prev_ring_pin = cur_ring_pin;
//		if (ring_detected) {
//			ESP_LOGI(TAG, "ring");
//		}
		
		// Evaluate our state
		switch (ring_state) {
			case RING_IDLE:
				if (ring_detected) {
					ring_state = RING_PULSE;
					ring_timer = RING_MAX_PULSE_MSEC / 10;
					ring_counts = 0;
					ring_pulse_counts = 1;
					ring_pulse_avg_sum = 0;
					prev_ring_usec = esp_timer_get_time();
				}
				break;
				
			case RING_PULSE:
				if (ring_detected) {
					ring_timer = RING_MAX_PULSE_MSEC / 10;
					ring_pulse_counts += 1;
					cur_ring_usec = esp_timer_get_time();
					ring_pulse_avg_sum += (cur_ring_usec - prev_ring_usec) / 1000;
					prev_ring_usec = cur_ring_usec;
				} else if (--ring_timer == 0) {
					if (ring_pulse_counts > 1) {
						ring_state = RING_WAIT;
						ring_counts += 1;
						ring_timer = RING_MAX_WAIT_MSEC / 10;
						
						if (ring_counts == 1) {
							// Notify audio_task to look for CID
							xTaskNotify(task_handle_cid, CID_NOTIFY_START_SAMPLE, eSetBits);
							cid_enabled = true;
						}
						
						ESP_LOGI(TAG, "Ring: %d pulses, avg %d mSec", ring_pulse_counts, ring_pulse_avg_sum / (ring_pulse_counts - 1));
					} else {
						ring_state = RING_IDLE;
						ESP_LOGE(TAG, "Spurious ring");
					}
				}
				break;
				
			case RING_WAIT:
				if (ring_detected) {
					ring_state = RING_PULSE;
					ring_timer = RING_MAX_PULSE_MSEC / 10;
					ring_pulse_counts = 1;
					ring_pulse_avg_sum = 0;
					if (cid_enabled) {
						cid_enabled = false;
						xTaskNotify(task_handle_cid, CID_NOTIFY_STOP_SAMPLE, eSetBits);
					}
					prev_ring_usec = esp_timer_get_time();
				} else if (--ring_timer == 0) {
					ring_state = RING_IDLE;
					
					if (cid_enabled) {
						cid_enabled = false;
						xTaskNotify(task_handle_cid, CID_NOTIFY_STOP_SAMPLE, eSetBits);
					}
					
					// Print ring info
					ESP_LOGI(TAG, "Ring total: %d", ring_counts);
				}
				break;
				
			default:
				ring_state = RING_IDLE;
		}
		
		vTaskDelay(pdMS_TO_TICKS(10));
	}
}
